"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DriverMessageHandler = void 0;
const error_1 = require("../error");
const command_1 = require("./command");
const state_1 = require("../state");
class DriverMessageHandler {
    remoteController;
    clientsController;
    logger;
    driver;
    client;
    constructor(remoteController, clientsController, logger, driver, client) {
        this.remoteController = remoteController;
        this.clientsController = clientsController;
        this.logger = logger;
        this.driver = driver;
        this.client = client;
    }
    async handle(message) {
        const { command } = message;
        switch (message.command) {
            case command_1.DriverCommand.getConfig: {
                const config = (0, state_1.dumpDriver)(this.driver, this.client.schemaVersion);
                return { config };
            }
            case command_1.DriverCommand.disableStatistics: {
                this.driver.disableStatistics();
                return {};
            }
            case command_1.DriverCommand.enableStatistics: {
                this.driver.enableStatistics({
                    applicationName: message.applicationName,
                    applicationVersion: message.applicationVersion,
                });
                return {};
            }
            case command_1.DriverCommand.getLogConfig: {
                const config = (0, state_1.dumpLogConfig)(this.driver, this.client.schemaVersion);
                return { config };
            }
            case command_1.DriverCommand.updateLogConfig: {
                this.driver.updateLogConfig(message.config);
                // If the logging event forwarder is enabled, we need to restart
                // it so that it picks up the new config.
                this.clientsController.restartLoggingEventForwarderIfNeeded();
                this.clientsController.clients.forEach((cl) => {
                    cl.sendEvent({
                        source: "driver",
                        event: "log config updated",
                        config: (0, state_1.dumpLogConfig)(this.driver, cl.schemaVersion),
                    });
                });
                return {};
            }
            case command_1.DriverCommand.isStatisticsEnabled: {
                const statisticsEnabled = this.driver.statisticsEnabled;
                return { statisticsEnabled };
            }
            case command_1.DriverCommand.startListeningLogs: {
                this.client.receiveLogs = true;
                this.clientsController.configureLoggingEventForwarder(message.filter);
                return {};
            }
            case command_1.DriverCommand.stopListeningLogs: {
                this.client.receiveLogs = false;
                this.clientsController.cleanupLoggingEventForwarder();
                return {};
            }
            case command_1.DriverCommand.checkForConfigUpdates: {
                const installedVersion = this.driver.configVersion;
                const newVersion = await this.driver.checkForConfigUpdates();
                const updateAvailable = newVersion !== undefined;
                return { installedVersion, updateAvailable, newVersion };
            }
            case command_1.DriverCommand.installConfigUpdate: {
                const success = await this.driver.installConfigUpdate();
                return { success };
            }
            case command_1.DriverCommand.setPreferredScales: {
                this.driver.setPreferredScales(message.scales);
                return {};
            }
            case command_1.DriverCommand.enableErrorReporting: {
                // This capability no longer exists but we keep the command here for backwards
                // compatibility.
                this.logger.warn("Z-Wave JS no longer supports enabling error reporting. If you are using " +
                    "an application that integrates with Z-Wave JS and you receive this " +
                    "error, you may need to update the application.");
                return {};
            }
            case command_1.DriverCommand.softReset: {
                await this.driver.softReset();
                return {};
            }
            case command_1.DriverCommand.trySoftReset: {
                await this.driver.trySoftReset();
                return {};
            }
            case command_1.DriverCommand.hardReset: {
                setTimeout(() => this.remoteController.hardResetController(), 1);
                return {};
            }
            case command_1.DriverCommand.shutdown: {
                const success = await this.driver.shutdown();
                return { success };
            }
            case command_1.DriverCommand.updateOptions: {
                this.driver.updateOptions(message.options);
                return {};
            }
            case command_1.DriverCommand.sendTestFrame: {
                const status = await this.driver.sendTestFrame(message.nodeId, message.powerlevel);
                return { status };
            }
            default: {
                throw new error_1.UnknownCommandError(command);
            }
        }
    }
}
exports.DriverMessageHandler = DriverMessageHandler;
