"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ZnifferMessageHandler = void 0;
const zwave_js_1 = require("zwave-js");
const error_1 = require("../error");
const command_1 = require("./command");
class ZnifferMessageHandler {
    driver;
    clientsController;
    zniffer;
    constructor(driver, clientsController) {
        this.driver = driver;
        this.clientsController = clientsController;
    }
    forwardEvent(data, minSchemaVersion = 38) {
        // Forward event to all clients
        this.clientsController.clients.forEach((client) => this.sendEvent(client, data, minSchemaVersion));
    }
    sendEvent(client, data, minSchemaVersion) {
        // Send event to connected client only
        if (client.receiveEvents &&
            client.isConnected &&
            client.schemaVersion >= (minSchemaVersion ?? 0)) {
            client.sendEvent(data);
        }
    }
    async handle(message) {
        const { command } = message;
        if (message.command != command_1.ZnifferCommand.init && !this.zniffer) {
            throw new Error("Zniffer is not running");
        }
        switch (message.command) {
            case command_1.ZnifferCommand.init: {
                if (this.zniffer) {
                    throw new Error("Zniffer is already running");
                }
                if (message.options.logConfig === undefined) {
                    message.options.logConfig = this.driver.options.logConfig;
                }
                if (message.options.securityKeys === undefined) {
                    message.options.securityKeys = this.driver.options.securityKeys;
                }
                if (message.options.securityKeysLongRange === undefined) {
                    message.options.securityKeysLongRange =
                        this.driver.options.securityKeysLongRange;
                }
                this.zniffer = new zwave_js_1.Zniffer(message.devicePath, message.options);
                this.zniffer
                    .on("ready", () => this.forwardEvent({
                    source: "zniffer",
                    event: "ready",
                }))
                    .on("corrupted frame", (corruptedFrame, rawDate) => this.forwardEvent({
                    source: "zniffer",
                    event: "corrupted frame",
                    corruptedFrame,
                    rawDate,
                }))
                    .on("frame", (frame, rawData) => this.forwardEvent({
                    source: "zniffer",
                    event: "frame",
                    frame,
                    rawData,
                }))
                    .on("error", (error) => this.forwardEvent({
                    source: "zniffer",
                    event: "error",
                    error,
                }));
                await this.zniffer.init();
                return {};
            }
            case command_1.ZnifferCommand.start: {
                await this.zniffer?.start();
                return {};
            }
            case command_1.ZnifferCommand.clearCapturedFrames: {
                this.zniffer?.clearCapturedFrames();
                return {};
            }
            case command_1.ZnifferCommand.getCaptureAsZLFBuffer: {
                return {
                    capture: Buffer.from(this.zniffer.getCaptureAsZLFBuffer().buffer),
                };
            }
            case command_1.ZnifferCommand.stop: {
                await this.zniffer?.stop();
                return {};
            }
            case command_1.ZnifferCommand.destroy: {
                this.zniffer?.removeAllListeners();
                await this.zniffer?.destroy();
                this.zniffer = undefined;
                return {};
            }
            case command_1.ZnifferCommand.capturedFrames: {
                return { capturedFrames: this.zniffer.capturedFrames };
            }
            case command_1.ZnifferCommand.supportedFrequencies: {
                return { frequencies: this.zniffer.supportedFrequencies };
            }
            case command_1.ZnifferCommand.currentFrequency: {
                return { frequency: this.zniffer.currentFrequency };
            }
            case command_1.ZnifferCommand.setFrequency: {
                await this.zniffer?.setFrequency(message.frequency);
                return {};
            }
            default: {
                throw new error_1.UnknownCommandError(command);
            }
        }
    }
}
exports.ZnifferMessageHandler = ZnifferMessageHandler;
